/*
 * OpenGLHIView.cpp
 *
 * Copyright (c) 2006, Arne Scheffler
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the
 * following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 * following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the Author nor the names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "OpenGLHIView.h"

//-------------------------------------------------------------------------------------------
OpenGLHIView::OpenGLHIView (HIViewRef inView)
: TView (inView)
, aglContext (0)
, pixelFormat (0)
, windowEventHandler (0)
, windowEventHandlerUPP (0)
, isResizing (false)
, opaque (false)
{
}

//-------------------------------------------------------------------------------------------
OpenGLHIView::~OpenGLHIView ()
{
	CleanupOpenGL ();
}

//-------------------------------------------------------------------------------------------
void OpenGLHIView::OwningWindowChanged (WindowRef oldWindow, WindowRef newWindow)
{
	TView::OwningWindowChanged (oldWindow, newWindow);
	CleanupOpenGL ();
	if (newWindow)
		CreateOpenGLContextForWindow (newWindow);
}

//-------------------------------------------------------------------------------------------
void OpenGLHIView::BoundsChanged (UInt32 inOptions, const HIRect& inOriginalBounds, const HIRect& inCurrentBounds)
{
	TView::BoundsChanged (inOptions, inOriginalBounds, inCurrentBounds);
	if (!isResizing)
		SetupOpenGLBounds ();
}

//-------------------------------------------------------------------------------------------
OSStatus OpenGLHIView::Initialize (TCarbonEvent& inEvent)
{
	inEvent.GetParameter('opaq', &opaque);
	return noErr;
}

//-------------------------------------------------------------------------------------------
void OpenGLHIView::Draw (RgnHandle inLimitRgn, CGContextRef inContext)
{
	aglSetCurrentContext (aglContext);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	DrawOpenGL ();

	aglSwapBuffers (aglContext);
}

//-------------------------------------------------------------------------------------------
void OpenGLHIView::SetupOpenGLBounds ()
{
	if (aglContext && GetWindowRef ())
	{
		aglSetCurrentContext (aglContext);
		aglUpdateContext (aglContext);

		HIViewRef contentView;
		HIViewFindByID (HIViewGetRoot (GetWindowRef ()), kHIViewWindowContentID, &contentView);
		HIRect r = Bounds ();
		HIRectConvert (&r, kHICoordSpaceView, GetViewRef (), kHICoordSpaceView, contentView);

		HIRect contentBounds;
		HIViewGetBounds (contentView, &contentBounds);

		GLint aiRect[4];
		aiRect[0] = r.origin.x;
		aiRect[1] = contentBounds.size.height - r.size.height - r.origin.y;
		aiRect[2] = r.size.width;
		aiRect[3] = r.size.height;
		
		aglSetInteger (aglContext, AGL_BUFFER_RECT, aiRect);
		aglEnable (aglContext, AGL_BUFFER_RECT);

		glViewport (0, 0, r.size.width, r.size.height);

		OpenGLBoundsChanged ();

		aglUpdateContext (aglContext);
	}
}

//-------------------------------------------------------------------------------------------
void OpenGLHIView::setOpaque (bool state)
{
	if (opaque != state)
	{
		opaque = state;
		if (aglContext)
		{
			aglSetCurrentContext (aglContext);
			GLint opacity = opaque ? 1 : 0;
			aglSetInteger (getAGLContext (), AGL_SURFACE_OPACITY, &opacity);
			aglUpdateContext (aglContext);
		}
	}
}

//-------------------------------------------------------------------------------------------
const GLint* OpenGLHIView::GetOpenGLAttributes ()
{
	static const GLint attrib[] = { AGL_RGBA, AGL_DOUBLEBUFFER, AGL_ACCELERATED, AGL_SAMPLE_BUFFERS_ARB, 1, AGL_SAMPLES_ARB, 2, AGL_NO_RECOVERY, AGL_DEPTH_SIZE, 16, AGL_NONE };
	return attrib;
}

//-------------------------------------------------------------------------------------------
OSStatus OpenGLHIView::CreateOpenGLContextForWindow (WindowRef window)
{
	CGrafPtr port = GetWindowPort (window);
	pixelFormat = aglChoosePixelFormat (NULL, 0, GetOpenGLAttributes ());
	aglContext = aglCreateContext (pixelFormat, NULL);
	aglSetDrawable (aglContext, port);
	aglSetCurrentContext (aglContext);
	aglUpdateContext (aglContext);
	glClearColor (0.0,0.0,1.0,1.0);

	GLint opacity = opaque ? 1 : 0;
	aglSetInteger (getAGLContext (), AGL_SURFACE_OPACITY, &opacity);

	SetupOpenGL ();

	SetupOpenGLBounds ();
	
	windowEventHandlerUPP = NewEventHandlerUPP (WindowEventHandler);

	const EventTypeSpec windowEvents[] = {
		{ kEventClassWindow, kEventWindowBoundsChanged },
	};
	
	InstallWindowEventHandler (window, windowEventHandlerUPP, GetEventTypeCount (windowEvents), windowEvents, this, &windowEventHandler);

	return noErr;
}

//-------------------------------------------------------------------------------------------
void OpenGLHIView::CleanupOpenGL ()
{
	if (aglContext)
	{
		aglSetDrawable (aglContext, NULL);
		aglSetCurrentContext (NULL);
		aglDestroyContext (aglContext);
		aglContext = 0;
	}
	if (pixelFormat)
	{
		aglDestroyPixelFormat (pixelFormat);
		pixelFormat = 0;
	}
	if (windowEventHandler)
		RemoveEventHandler (windowEventHandler);
	if (windowEventHandlerUPP)
		DisposeEventHandlerUPP (windowEventHandlerUPP);
}

//-------------------------------------------------------------------------------------------
OSStatus OpenGLHIView::WindowEventHandler (EventHandlerCallRef inHandler, EventRef inEvent, void* inUserData)
{
	OpenGLHIView* glHiView = (OpenGLHIView*)inUserData;
	if (GetEventClass (inEvent) == kEventClassWindow)
	{
		if (GetEventKind (inEvent) == kEventWindowBoundsChanged)
		{
			TCarbonEvent event (inEvent);
			UInt32 attrs;
			event.GetParameter (kEventParamAttributes, &attrs);
			if (attrs & kControlBoundsChangeSizeChanged)
			{
				glHiView->isResizing = true;
				HIViewSetDrawingEnabled (glHiView->GetViewRef (), false);
				OSStatus result = CallNextEventHandler (inHandler,inEvent);
				glHiView->SetupOpenGLBounds ();
				HIViewSetDrawingEnabled (glHiView->GetViewRef (), true);
				HIViewSetNeedsDisplay(glHiView->GetViewRef (), true);
				glHiView->isResizing = false;
				return result;
			}
		}
	}
	return eventNotHandledErr;
}

